/*
 * Copyright (c) 2000-2002 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#ifndef RaGator_h
#define RaGator_h

#include <stdlib.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>


#define RA_MODELTAGSTR		"Model"
#define RA_FLOWTAGSTR		"Flow"

#define RA_MODELIST		1
#define RA_FLOWLIST		2

#define RA_FLOWPOLICYFIELDNUM	9
#define RA_MODELPOLICYFIELDNUM	7
  
#define RA_LABELSTRING		0
#define RA_POLICYID		1
#define RA_POLICYSRCADDR	2
#define RA_POLICYDSTADDR	3
#define RA_POLICYPROTO		4
#define RA_POLICYSRCPORT	5
#define RA_POLICYDSTPORT	6
#define RA_POLICYMODELST	7
#define RA_POLICYTIMEOUT	8

#define RA_MODIFIED		1

#define RA_CON			1
#define RA_DONE			2

#define RA_CIDR_TYPE		1
#define RA_NETCLASS_TYPE	2

 
#include <netinet/in.h>

#include <ctype.h>
#include <argus-namedb.h>

#include <argus_filter.h>
 

struct ArgusQueueHeader {
   struct ArgusQueueHeader *prv, *nxt;
   struct RaQueueStruct *queue;
   int status;
   struct timeval lasttime, logtime;
};

struct RaQueueStruct {
   int count;
   struct ArgusQueueHeader *start;
};

struct ArgusTimeStatsStruct {
   int n;
   long long sumtime;
   double sumsqrd;
};

struct ArgusRecordData {
   int status, farhdrstatus;
   struct ArgusFarHeaderStruct *farhdrs[32];
   struct ArgusFlow flow;
   struct ArgusRecord *argus;
   struct ArgusAGRStruct agr;
   struct ArgusTimeStatsStruct act, idle;
};

struct ArgusRecordStore {
   struct ArgusQueueHeader qhdr;
   struct RaHashTableHeader *rahtblhdr;
   int  status, ArgusTimeout;
   struct ArgusRecordData *data;
};


char *RaPolicyError [] = {
   "Malformed String",
   NULL,
};

struct RaCIDRAddr {
   u_int type, addr, mask;
};

struct RaPolicyStruct {
   u_int RaEntryType, RaPolicyId;
   struct RaCIDRAddr src, dst;
   u_char proto;
   u_short sport, dport;
   u_int RaModelId, ArgusTimeout;
   char *str;
};


struct RaHashTableHeader {
   struct RaHashTableHeader *nxt, *prv;
   unsigned short hash;
   struct ArgusFlow flow;
   struct ArgusRecordStore *storeobj;
};

struct RaHashTableStruct {
   int size, count;
   struct RaHashTableHeader **array;
};



struct ArgusListObjectStruct {
   struct ArgusListObjectStruct *nxt, *prv;
   void *obj;
};


struct ArgusListStruct {
   struct ArgusListObjectStruct *start;
   unsigned int count;
};


struct RaCIDRAddr RaCIDRBuffer, *RaCIDRPtr = &RaCIDRBuffer;

int RaThisFlowNum = 0;
int RaThisModelNum = 0;
int RaParseError = 0;


struct RaHashTableStruct RaHashTable;

struct RaPolicyStruct  *RaFlowArray[1024];
struct RaPolicyStruct *RaModelArray[1024];

struct RaPolicyStruct *RaFlowModelPolicy = NULL;

struct RaQueueStruct *RaArgusRecordQueue = NULL;
struct RaQueueStruct *RaModelerQueue = NULL;

int RaPolicyMatch (struct ArgusRecord *, struct RaPolicyStruct *);

int RaCIDRAddrMatches (u_int, struct RaCIDRAddr *);
int RaProtoMatches (u_short, int);
int RaPortMatches (u_short p1, u_short p2);

unsigned int RaIndexArgusRecord (struct ArgusRecord *, struct ArgusFarHeaderStruct **);

struct RaPolicyStruct *RaFlowModelOverRides (struct ArgusRecord *);
void RaModifyFlow(struct RaPolicyStruct *, struct ArgusRecord *);
void RaModifyDefaultFlow(struct ArgusRecord *);

struct RaHashTableHeader *RaFindHashObject (void);
struct RaHashTableHeader *RaAddHashEntry (struct ArgusRecordStore *);
void RaRemoveHashEntry (struct RaHashTableHeader *);

void RaSendArgusRecord(struct ArgusRecordStore *);


struct RaQueueStruct *RaNewQueue (void);
void RaDeleteQueue (struct RaQueueStruct *);

int RaGetQueueCount(struct RaQueueStruct *);

void RaPushQueue(struct RaQueueStruct *, struct ArgusQueueHeader *);
int RaAddToQueue(struct RaQueueStruct *, struct ArgusQueueHeader *);

struct ArgusQueueHeader *RaPopQueue(struct RaQueueStruct *);
struct ArgusQueueHeader *RaRemoveFromQueue(struct RaQueueStruct *, struct ArgusQueueHeader *);

void RaEmptyQueue(struct RaQueueStruct *);

struct ArgusListStruct *ArgusNewList(void);
void ArgusDeleteList(struct ArgusListStruct *);
int ArgusListEmpty (struct ArgusListStruct *);
int ArgusGetListCount (struct ArgusListStruct *);
void *ArgusFrontList (struct ArgusListStruct *);
void ArgusPushFrontList(struct ArgusListStruct *, void *);
void ArgusPopFrontList (struct ArgusListStruct *);
void ArgusPushBackList (struct ArgusListStruct *, void *);
void ArgusPopBackList  (struct ArgusListStruct *);

int RaCheckTimeout(struct ArgusRecordStore *obj, struct ArgusRecord *);

void RaProcessQueue(struct RaQueueStruct *, unsigned char state);
void RaEmptyQueue(struct RaQueueStruct *);

struct ArgusRecordStore *RaFindArgusRecord (struct ArgusRecord *);
struct ArgusRecordStore *RaNewArgusStore (struct ArgusRecord *);
struct ArgusRecordData *RaNewArgusData (struct ArgusRecord *);
void RaRemoveArgusRecord (struct ArgusRecord *);

void RaMergeArgusRecord (struct ArgusRecord *, struct ArgusRecordData *);

void RaUpdateArgusStore(struct ArgusRecord *, struct ArgusRecordStore *);
void RaTimeoutArgusStore (struct ArgusRecordStore *);

int RaCreatePolicyEntry (int, char *);
int RaReadFlowModelFile (char **);

struct RaPolicyStruct *RaParsePolicyEntry (char *);

struct RaCIDRAddr *RaParseCIDRAddr (char *);

#endif
